*:************************************************************
*:
*:  Source File: tbprint.prg
*:
*:       System: TBPrint - print TBrowse
*:       Authors: Kevin Farley and David Barrett, 
*:                expanding on code by Pat Tormey and Les Squires
*:
*:   Placed in the public domain.
*:
*:  Comments : Routines for TBPrint - print current TBrowse
*:
*:  Procs & Funcs: TBPrint()
*:               : PrintColumnHead()
*:               : getColumnHead()
*:               : GetColumnLen()
*:               : HeadLen()
*:               : X2Char()
*:
*: Documented 06/30/92 at 12:23:02 DOC Vers 2.10, Source Code Documenter
*:************************************************************
/*
   TBPrint will print the current tBrowse extracting all the information
   it needs from the tbrowse object.  Optionally, if passed the number of
   lines per page, it will insert page breaks in the printed output. 
   It also assumes that SET PRINTER TO has been used to set the printer 
   to the device or print file which is to receive the printed output if 
   the output is not to go to LST.

   These routines are expanded from and includes parts of code posted by
   Pat Tormey in BCSTBPN.PRG and include ideas by Les Squires and others
   in a thread on Nanforum about printing a tBrowse.  This routine
   determines the column width by using TBColumn:width if it has been set,
   otherwise is uses the widest of the either the footing, heading or the
   data return from the first eval of TBColumn:block.  Determining the
   width of the heading and footing takes into account any embedded semi-
   colons.

*/

#command DEFAULT <p> TO <val> [,<pn> TO <valn>] ;
         => ;
         <p> = if(<p> = NIL, <val>, <p>) ;
         [;<pn> = if(<pn> = NIL, <valn>, <pn>)]

#define FF chr( 12 )
#define CRLF chr( 13 ) + chr( 10 )
#include "inkey.ch"
#include "box.ch"
#define COLSEP '   '

*!************************************************************
*!
*!             FUNCTION TBPrint
*!
*!     Calls : GetColumnHead()
*!           : GetColumnLen()
*!           : PrintColumnHead()
*!           : X2Char()
*!
*!   Purpose : Send current TBrowse to printer, optionally with page
*!             breaks and headers.
*!
*!   In      : oB - TBrowse object
*!           : nLinesPerPage - if 0 cause page breaks to be inserted
*!                             after this many lines. Defaults to 0.
*!           : cTitle - if nLinesPerPage not 0 printed at top of
*!                      each page along with page #. Defaults to ''.
*!
*!   Out     : .F. if user aborted with ESC key, otherwise .T.
*!
*!   Assumes : a TBrowse has a goTop() method
*!
*!************************************************************
function TBPrint( oB, nLinesPerPage, cTitle )
  local i := 0, m ,               ;  // counters
        lAbort := .F. ,           ;
        nPageNo := 0 ,            ;
        cPageHeading := padc(cTitle,65) + " Page No: ",      ;  
        aColumn ,                 ;  // temp array
        aColHeading := {} ,       ;  // array holding column headings
        nColRows := 0 ,           ;  // number of print rows in headings
        nRecord := recno() ,      ;  // record pointer place to restore
        nLinesThisPage ,          ;  // lines printed current page
        aLen := {} ,              ;  // column widths stored in this array
        cScreen                      // screen snapshot

  DEFAULT nLinesPerPage TO 0
  DEFAULT cTitle TO ''

  set print on
  set console off

  // find column widths, store in aLen array
  aLen := GetColumnLen( ob )

  // make an array of column headers, store a aColHeading array
  for i := 1 to oB:colCount()
    aColumn := GetColumnHead( ob:getcolumn(i):Heading(), aLen[ i ] )
    aadd( aColHeading, aColumn )
    nColRows := max( nColRows, len( aColumn ) )
  next

  // For Progress feedback
  cScreen = savescreen( 10, 10, 16, 71 )
  @ 10, 10, 15, 70 BOX B_DOUBLE + " "

  // if pages breaks, print title and page # on first page
  if nLinesPerPage > 0
    Pager( cPageHeading, ++ nPageNo )
  endif

  // print column headings from aColHeading array
  nLinesThisPage := PrintColumnHead( aColHeading, nColRows, oB, aLen )

  ob:goTop()

  while ( ! ob:hitBottom ) .and. ( ! lAbort )

    // For Progress Feedback
    if nLinesPerPage > 0
       Center( 12, "Currently printing . . . Page: " + ;
               ltrim(str(nPageNo)) + "  Line: " + ltrim(str(nLinesThisPage)) )
    else
       Center( 12, "Currently printing . . .  Line: " + ltrim(str(nLinesThisPage)) )
    endif

    for i := 1 to Ob:colCount()
      // first Get the Tb Column, then get the Column's Block
      // evaluate it, then Convert it to a String, and add it to the Array
      ?? padc( X2Char(eval(ob:getcolumn(i):Block())), aLen[ i ] )  // getting cryptic Huh?
      ?? if( i < Ob:colCount(), COLSEP, '' )
    next

    ?? CRLF

    oB:hitBottom := ( 0 == eval(oB:skipBlock, 1) )

    nLinesThisPage ++

    // if page breaks enabled and this page full, print page break
    if nLinesPerPage > 0 .and. ( nLinesThisPage >= nLinesPerPage )
      Pager( cPageHeading, ++ nPageNo )
      nLinesThisPage := PrintColumnHead( aColHeading, nColRows, oB, aLen )
    endif

    // if user pressed ESC key ask if want to abort, abort if yes
    if inkey() == K_ESC
      set print off
      set console on
      if YesOrNo( 'Abandon printing?' )
        lAbort := .T.
      endif
      set print on
      set console off
    endif

  enddo

  // For Progress Feedback
  restscreen( 10, 10, 16, 71, cScreen )

  // Reset everything
  oB:goTop()
  oB:refreshAll()
  oB:stabilize()

  set print off
  set console on

return ! lAbort


*!************************************************************
*!
*!             FUNCTION Pager
*!
*! Called by : TBPrint
*!
*!   Purpose : Print Page header
*!
*!   In      : cHeader - Page Header
*!             nPage   - Page #
*!
*!   Out     : NIL
*!
*!   Assumes :
*!
*!************************************************************
static function Pager( cHeader, nPage )
   DEFAULT cHeader TO space(65),  nPage TO 1
   qqout(FF + cHeader + ltrim( str( nPage ) ) + CRLF + CRLF )
return NIL


*!************************************************************
*!
*!             FUNCTION PrintColumnHead
*!
*! Called by : TBPrint
*!
*!   Purpose : Print column headings contained in array.
*!
*!   In      : aColHeading - array containing column headings.
*!           : nColRows - number of print rows contained in aColHeading.
*!           : oB - tBrowse object.
*!           : aLen - array containing column widths.
*!
*!   Out     : print row for current page
*!
*!   Assumes :
*!
*!************************************************************
static function PrintColumnHead( aColHeading, nColRows, oB, aLen )
  local m, i   // counters
  for m := 1 to nColRows
    for i := 1 to Ob:colCount()
      ?? if( len( aColHeading[ i ] ) >= m, aColHeading[ i, m ], ;
                              space( aLen[ i ] + len( COLSEP ) ) )
    next
    ?? CRLF
  next
return 2 + nColRows


*!************************************************************
*!
*!             FUNCTION getColumnHead
*!
*! Called by : TBPrint
*!
*!   Purpose : Create array for column header, storing seperate lines
*!             of header (parts seperated by semicolons) in seperate
*!             array elements, centering based on column width
*!
*!   In      : cHeading - column header which may contain semicolons.
*!           : nColWidth - width for column, to be used for centering.
*!
*!   Out     : array for this column heading.
*!
*!   Assumes : nothing
*!
*!************************************************************
static function getColumnHead( cHeading, nColWidth )
  local aColumn := {} ,     ;     // array to be filled out and returned.
          cTemp := '' ,     ;     // temp. holder for each header element.
          m := 0 ,          ;     // counter
          cChar                   // temp. holder for each character.
  do while ++m <= len( cHeading )
    cChar := substr( cHeading, m, 1 )
    if cChar == ";"
      aadd( aColumn, padc( cTemp, nColWidth + len( COLSEP ) ) )
      cTemp := ''
    else
      cTemp += cChar
    endif
  enddo
  aadd( aColumn, padc( cTemp, nColWidth + len( COLSEP ) ) )
return aColumn

*!************************************************************
*!
*!             FUNCTION GetColumnLen
*!
*! Called by : TBPrint
*!
*!     Calls : HeadLen()
*!           : X2Char()
*!
*!   Purpose : Create array containing column widths for tbrowse,
*!             determined by looking at column width, header, footer
*!             and block length.
*!
*!   In      : ob - tbrowse object
*!
*!   Out     : aLen - array containing width of each column in tbrowse.
*!
*!   Assumes : nothing
*!
*!************************************************************
static function GetColumnLen( ob )
  local i ,         ;    // loop counter
        nThisLen ,  ;    // variable to hold current column width.
        aLen := {}       // array of widths to return.
  for i := 1 to Ob:colCount()
    // get the Column Width
    if ob:getcolumn(i):width == NIL
      // if column width not set by user column width is largest of
      //  len of heading, len of footing, or len of eval( block )
      nThisLen := max( len( X2Char( eval( ob:getcolumn(i):Block() ))),     ;
                       max( HeadLen( ob:getcolumn(i):Heading()  ) ,        ;
                            HeadLen( ob:getcolumn(i):Footing()  )          ;
                          )                                                ;
                     )
    else
      nThisLen := ob:getcolumn(i):width
    endif
    aadd(aLen, nThisLen )
  next
return aLen

*!************************************************************
*!
*!             FUNCTION HeadLen
*!
*! Called by : GetColumnLen
*!
*!   Purpose : Returns max width of header which may contain semicolons.
*!
*!   In      : cHeading - header to determine max length of.
*!
*!   Out     : nMaxLen - maximum length of cHeading
*!
*!   Assumes : nothing
*!
*!************************************************************
static function HeadLen( cHeading )
  local nLen := 0,         ;     // current segment width.
        nMaxLen := 0,      ;     // holder of max width found so far
        m := 0,            ;     // loop counter
        cChar                    // current character
  DEFAULT cHeading TO ''
  do while ++m <= len( cHeading )
    cChar := substr( cHeading, m, 1 )
    if cChar == ';'
      nMaxLen := max( nMaxLen, nLen )
      nLen := 0
    else
      nLen ++
    endif
  enddo
return nMaxLen


*!************************************************************
*!
*!             FUNCTION X2Char
*!
*! Called by : TBPrint
*!           : GetColumnLen
*!
*!   Purpose : Convert anything to character.
*!
*!   In      : X - variable of undetermined type
*!
*!   Out     : cX - character representation of X.
*!
*!   Assumes :
*!
*!************************************************************
static function X2Char( X )
  // convert everything to Character
  local cX := ''
  do case      // only need to do this IF aIndex hasn't been made yet
  case valtype( X ) = "C"
    cX := X
  case valtype( X ) = "N"
    cX := str( X )
  case valtype( X ) = "D"
    cX := dtoc( X )
  case valtype( X ) = "L"
    cX := if( X, 'YES', 'NO ' )
  case valtype( X ) = "A"
    cX := "ARRAY!"
  case valtype( X ) = "M"
    cX := "MEMO"
  otherwise   // just in case
    cX := str( recno() )
  endcase

return cX

function YesOrNo( cMessage )
  local cScreen, cYesOrNo:= ' ', GetList := {}
  cScreen = savescreen( 10, 10, 16, 71 )
  @ 10, 10, 15, 70 BOX B_DOUBLE + " "
  Center( 12, cMessage )
  set cursor on
  @ 13, 25 SAY 'Press Y(es) or N(o) ' GET cYesOrNo VALID (cYesOrNo $ 'YyNn')
  read
  set cursor off
  restscreen( 10, 10, 16, 71, cScreen )
return if( ( cYesOrNo $ 'Yy' ), .T., .F. )

function Center( nRow, cMessage )
  local nY
  nY := INT ( ( maxcol() - len( cMessage ) ) / 2 )
  @ nRow, nY SAY cMessage
return nY            

*:   EOF: tbprint.prg
